import { App } from "./components/App.tsx";
import "./index.css";
import "./monaco.ts";
import { StrictMode } from "react";
import { createRoot } from "react-dom/client";

createRoot(document.getElementById("root")!).render(
  <StrictMode>
    <App />
  </StrictMode>,
);
