.App {
}

.App_SidePane {
  display: grid;
  grid-template:
    "TabPanel" minmax(0, 1fr)
    "RunPanel" auto;
  background-color: var(--panel-background-color);
  height: 100vh;
}

.App_Tabs {
  display: grid;
  grid-area: TabPanel;
  grid-template:
    "Tabs" auto
    "TabPanel" minmax(0, 1fr);
}

.App_TabList {
  display: flex;
  grid-area: Tabs;
  margin: 0;
  background-color: var(--panel-tabs-background-color);
  padding: 0.5rem 1rem 0;
}

.App_Tab {
  display: block;
  transition:
    background-color 0.2s,
    border-color 0.2s;
  cursor: pointer;
  outline: none;
  padding: 0.3rem 1rem;
  color: var(--input-text-color);
  font-size: 0.8rem;
  line-height: 1rem;
  user-select: none;
}

.App_Tab:hover {
  background-color: var(--input-background-color);
  color: var(--input-active-text-color);
}

.App_Tab__Spacer {
  cursor: auto;
  background: none;
  padding: 0;
}

.App_Tab__Link {
  text-decoration: none;
}

.App_Tab__Version {
  color: var(--code-color-escape);
}

.App_Tab__Version:hover {
  background-color: inherit;
  color: var(--code-color-escape);
  cursor: default;
}

.App_Tab__Link::after {
  vertical-align: middle;
  margin-left: 0.25rem;
  content: url("data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' width='12' height='12' fill='none' stroke='%23607d8b' stroke-linecap='round' stroke-linejoin='round' stroke-width='1.5' viewBox='0 0 24 24'%3E%3Cpath stroke='none' d='M0 0h24v24H0z'/%3E%3Cpath d='M11 7H6a2 2 0 0 0-2 2v9a2 2 0 0 0 2 2h9a2 2 0 0 0 2-2v-5M10 14 20 4M15 4h5v5'/%3E%3C/svg%3E");
  color: var(--panel-elem-text-color);
}

.App_Tab__Icon {
  display: flex;
  margin-left: auto;
  border: none;
  background: none;
  padding: 0;
  width: 1.25rem;
  height: 1.25rem;
}

.App_Tab__Icon svg {
  width: 1.25rem;
  height: 1.25rem;
}

.App_Tab__Icon:hover {
  background: none;
}

.App_Tab__Active {
  background-color: var(--panel-background-color);
}

.App_TabPanel {
  display: none;
  grid-area: TabPanel;
  overflow: auto;
}

.App_TabPanel__NoScroll {
  overflow: hidden;
}

.App_TabPanel__Active {
  display: block;
}

.App_RunPanel {
  display: flex;
  grid-area: RunPanel;
  justify-content: end;
  align-items: center;
  gap: 1rem;
  border-top: 1px solid var(--panel-separator);
  padding: 0 2rem;
  height: 40px;
}

.App_RunPanelInputGroup {
  display: flex;
  gap: 0.3rem;
}

.App_MainPane {
  width: 100%;
  height: 100vh;
}

.App_MainPaneContent {
  padding: 2rem 4rem;
  width: 100%;
  height: 100vh;
  overflow: auto;
}
