.DiagramSettings {
  padding: 1rem 2rem;
}

.DiagramSettings_Group {
  display: grid;
  grid-template-columns: repeat(auto-fit, minmax(390px, 1fr));
  gap: 0.3rem 2rem;
}

.DiagramSettings_Item {
  display: grid;
  grid-template-columns: auto max-content max-content;
  gap: 0.3rem;
}

.DiagramSettings_Input {
  width: 150px;
}

.DiagramSettings_Checkbox {
  margin: 0;
  width: calc(1.6rem + 2px);
  height: calc(1.6rem + 2px);
}

.DiagramSettings_More {
  margin: 0.5rem 0;
}

.DiagramSettings_MoreToggle {
  cursor: pointer;
  font-size: 14px;
  user-select: none;
}

.DiagramSettings_MoreToggle::before {
  display: inline-block;
  width: 13px;
  content: "";
  color: var(--panel-elem-text-color);
  font-size: 10px;
}

.DiagramSettings_MoreToggle:hover::before {
  color: var(--panel-elem-hover-text-color);
}

.DiagramSettings_MoreToggle__Active::before {
  content: "▼";
}

.DiagramSettings_MoreToggle__Hidden::before {
  content: "▶";
}

.DiagramSettings_MoreContent {
  transition: 0.1s ease-in-out;
  overflow: hidden;
}

.DiagramSettings_MoreContent__Active {
  visibility: visible;
  opacity: 1;
}

.DiagramSettings_MoreContent__Hidden {
  visibility: hidden;
  opacity: 0;
  height: 0;
}

.DiagramSettings_Help {
  position: relative;
}

.DiagramSettings_HelpToggle {
  display: inline-block;
  width: calc(1.6rem + 2px);
  text-align: center;
}

.DiagramSettings_HelpMessage {
  position: absolute;
  top: 1.5rem;
  right: 0;
  transition: 0.1s ease-in-out;
  box-shadow: 1px 1px 3px var(--popup-shadow-color);
  border: 1px solid var(--popup-border-color);
  border-radius: var(--popup-border-radius);
  background-color: var(--popup-background-color);
  padding: 0.3rem 0.5rem;
  width: 300px;
  font-size: 0.8rem;
  text-align: left;
}

.DiagramSettings_HelpMessage__Active {
  visibility: visible;
  opacity: 1;
  z-index: 1;
}

.DiagramSettings_HelpMessage__Hidden {
  visibility: hidden;
  opacity: 0;
}
